#!/usr/bin/env python

'''Provides a RTP Proxy server to be used in conjunction with SER's mediaproxy
module. MediaProxy stands in between the 2 SIP clients participating in a VoIP
conversation and forwards the RTP and RTCP streams between them, allowing
the SIP clients to work behind NAT without special settings on their behalf.
MediaProxy must have a public IP.
'''

## mangle path depending on version (for binary package). also add modules subdir
import sys, os
version = '%d.%d' % sys.version_info[0:2]
sys.path[1:1] = [os.path.normpath((sys.path[0] or '.') + '/' + version)]
sys.path[0:0] = [os.path.normpath((sys.path[0] or '.') + '/modules')]
del os

if __name__ == '__main__':
    import rtphandler
    import daemon
    import sys
    from optparse import OptionParser
    from options import ExtendedOptions

    parser = OptionParser(option_class=ExtendedOptions, version="%prog 1.0")
    parser.add_option("--no-fork", action="store_false", dest="fork", default=1,
                      help="run the process in the foreground")
    parser.add_option("--ip", dest="proxyIP", type="ip", default=None,
                      help="use this IP to talk to the RTP parties (first found)",
                      metavar="IP")
    parser.add_option("--ports", dest="ports", type='ports', default=(35000, 65000),
                      help="rtp port range (35000:65000)",
                      metavar="min:max")
    parser.add_option("--socket", dest="controlSock", default='/var/run/mediaproxy.sock',
                      help="control socket or 'None' (/var/run/mediaproxy.sock)", 
                      metavar="File")
    parser.add_option("--group", dest="group", default='ser',
                      help="allow this group to write to the control socket (ser)", 
                      metavar="Group")
    parser.add_option("--listen", dest="listenAddr", type='netaddr', default=None,
                      help="address to listen on for remote control (None)", 
                      metavar="Host[:Port]")
    parser.add_option("--allow", dest="allowedHosts", type='list', default=['127.0.0.1'],
                      help="who is allowed to control this mediaproxy (127.0.0.1)", 
                      metavar="Host[,...]")
    parser.add_option("--expire", dest="timeout", type="int", default=60,
                      help="expire idle sessions after this many seconds (60)",
                      metavar="Secs")
    parser.add_option("--tos", dest="tos", type="hexdec", default=0xb8,
                      help="set TOS value on IP packets (0xb8)",
                      metavar="TOS")
    parser.add_option("--pid", dest="pidFile", default='/var/run/mediaproxy.pid',
                      help="pid file (/var/run/mediaproxy.pid)", 
                      metavar="File")
    parser.add_option("--log", dest="logFile", default=None,
                      help="log to this file too, along with syslog (None)",
                      metavar="File")
    (options, args) = parser.parse_args()

    ## Pass the command line options (or the default values) to the rtphandler module

    proxyIP = options.proxyIP or rtphandler.getHostIP()
    if not proxyIP:
        print "Cannot determine the IP address of this host! Use --ip to specify one."
        sys.exit(1)
    rtphandler.controlSock = options.controlSock
    rtphandler.controlSockGroup = options.group
    if options.listenAddr is not None:
        rtphandler.tcpListenAddress = options.listenAddr
    rtphandler.proxyIP = proxyIP
    rtphandler.minPort = options.ports[0]
    rtphandler.maxPort = options.ports[1]
    rtphandler.TOS = options.tos
    rtphandler.sessionTimeout = options.timeout
    pidFile = options.pidFile

    if options.fork:
        try:
            d = daemon.Daemon('mediaproxy', pidFile, logFile=options.logFile, useSyslog=1)
        except daemon.DaemonError, why:
            print >>sys.stderr, "Mediaproxy error: %s. Exiting." % why
            sys.exit(1)
    try:
        ## Start local command handler if enabled
        handler = None
        if options.controlSock.lower() != 'none':
            handler = rtphandler.ControlSocket(options.controlSock, options.group)
        where = handler and ("`%s'" % options.controlSock) or "is disabled"
        print "Listening for commands on local socket %s" % where

        ## Start remote command handler if requested
        if options.listenAddr is not None:
            allowed = options.allowedHosts
            if allowed and '127.0.0.1' not in allowed:
                allowed.append('127.0.0.1')
            handler = rtphandler.RemoteControl(options.listenAddr, allowed)
            print "Listening for remote commands on `%s:%s'" % options.listenAddr
            where = handler.allowedHostList and ", ".join(allowed) or "anywhere!!! beware!"
            print "Remote commands are allowed from %s" % where
        else:
            print "Listening for remote commands is disabled"

        ## Print some more startup info
        if handler:
            print "Using IP address `%s' for the RTP/RTCP proxy" % proxyIP
        else:
            print "No control sockets defined! Cannot receive any command at all. Exiting."

        ## Start main loop. Note that this will return immediately and exit if no control
        ## sockets are defined at all. this is why there is no sys.exit() above
        rtphandler.loop()

    except rtphandler.MediaProxyError, why:
        print >>sys.stderr, "Mediaproxy error: %s." % why
    except KeyboardInterrupt:
        print >>sys.stderr, "Got keyboard interrupt. Exiting."
    except daemon.DaemonQuit, msg:
        print >>sys.stderr, msg

    # ev call this
    #rtphandler.asyncore.close_all()

