
'''
 Implementation of a mediaproxy request dispatcher for SER.

 The dispatcher implements an architecture where SER will work with multiple
mediaproxy servers (local or remote). The dispatcher is implemented around the
concept of choosing which mediaproxy server to use based on the domain of the
sender or destination. If no mediaproxy server is defined for the given domain
it will fallback to a default mediaproxy server (if configured to use one).

 Mediaproxy servers for a domain are defined using DNS SRV records.
Multiple mediaproxy servers can be defined for a domain by using the priority
and weight attributes of the SRV records. This permits both load balancing the
traffic between multiple mediaproxy servers (if they are defined at the same
priority but with different weights) or mediaproxy server fallbacks (if none
of the servers defined at a priority responds, it will fallback to servers
defined at a lower priority).


Copyright 2004 Dan Pascu
'''

from SocketServer import ThreadingUnixStreamServer, BaseRequestHandler
import socket
import sys, os, grp, atexit, errno
import time
import threading
import random


## Script settings. Don't edit here. Use command line options to overwrite them

controlSock      = '/var/run/proxydispatcher.sock'
controlSockGroup = 'ser'

## This ca be a path (unix socket), TCP/IP address (host, port) or None
defaultMediaProxy = '/var/run/mediaproxy.sock'

## If this is defined (is not None) the dispatcher will use the mediaproxy
## servers defined by this domain as default mediaproxies.
## The value of defaultMediaProxy defined above will be ignored.
defaultDomain = None

## Internal state of dispatcher (DON'T TOUCH!!!)

Sessions = {}

## Sessions above should only be accessed after aquiring this lock
sessionLock = threading.RLock()

sessionTimeout = 24*60*60 ## remove stale sessions after 24 hours

maxDataSize = 8192

## Fix the environment

## Add path to foreign python modules that we include with mediaproxy for
## convenience since they are not standard. Some are also needed for python < 2.3
sys.path[1:1] = [os.path.normpath((sys.path[0] or '.') + '/external')]

import DNS
DNS.DiscoverNameServers()

## check if we have python 2.3 or newer. Else use some 3rd party module for timeout
## to work with sockets
try:
    socket.setdefaulttimeout
except:
    import timeoutsocket
    socket.setdefaulttimeout = timeoutsocket.setDefaultSocketTimeout
    socket.getdefaulttimeout = timeoutsocket.getDefaultSocketTimeout
    socket.timeout = timeoutsocket.Timeout
    del(timeoutsocket)

##
## Main loop (convenience function)
##

def run(server, timeout=sessionTimeout):
    timer = SessionTimer(timeout)
    while 1:
        server.serve_forever()
        timer.run()

## Helper functions

def warning(msg):
    print >>sys.stderr, "warning: " + str(msg)

def error(msg):
    print >>sys.stderr, "error: " + str(msg)

## Accessing the sessions thread safe
## This may be really fucked up and possibly buggy
## (my brain is no longer working since yesterday, so I really don't know anymore)
## Traiasca stilul heirupist

def getSession(sessionId):
    sessionLock.acquire()
    try:
        return Sessions[sessionId]
    finally:
        sessionLock.release()

def allSessions():
    sessionLock.acquire()
    try:
        return Sessions.values()
    finally:
        sessionLock.release()

## Classes

class CommandError(Exception):    pass
class DispatcherError(Exception): pass


class SessionTimer:
    def __init__(self, period):
        self.period = period
        self.next = time.time() + period

    def run(self):
        now = time.time()
        if now >= self.next:
            ## Check sessions and timeout stale ones
            count = 0
            sessionLock.acquire()
            try:
                for session in Sessions.values():
                    if now >= session.expire:
                        try:
                            del Sessions[session._id]
                        except KeyError:
                            pass
                        count += 1
            finally:
                sessionLock.release()
            now = time.time()
            self.next = now + self.period
            print "info: expired %d stale session(s)" % count
        return self.next - now


class SERCommandHandler(ThreadingUnixStreamServer):
    def __init__(self, path=None, group=None):
        ## If we would have passed these as default values they would have been
        ## locked to the values they had when the rtphandler module is loaded
        ## This way we can change the default values after rtphandler is loaded
        ## and we get the updated values
        path = path or controlSock
        group = group or controlSockGroup
        self.path = path
        self.group = group
        try:
            os.unlink(path)
        except OSError:
            pass
        try:
            ThreadingUnixStreamServer.__init__(self, path, SERRequestHandler)
        except socket.error, why:
            raise DispatcherError, "couldn't create server (%s)" % why[1]
        atexit.register(self.handle_close)

    def server_bind(self):
        ThreadingUnixStreamServer.server_bind(self)
        path = self.path
        try:
            groupId = grp.getgrnam(self.group)[2]
        except KeyError, IndexError:
            ## Make it world writable
            try:
                os.chmod(path, 0777)
            except OSError:
                warning("couldn't set access rights for %s." % path)
                warning("SER may be unable to communicate with us!")
        else:
            ## Make it writable only to the SIP proxy group members
            try:
                os.chown(path, -1, groupId)
                os.chmod(path, 0770)
                #os.chmod(path, S_IRUSR|S_IWUSR|S_IXUSR|S_IRGRP|S_IWGRP|S_IXGRP)
            except OSError:
                warning("couldn't set access rights for %s." % path)
                warning("SER may be unable to communicate with us!")

    def handle_close(self):
        self.socket.close()
        try:
            os.unlink(self.path)
        except OSError:
            pass


class SERRequestHandler(BaseRequestHandler):
    '''Handles a command received through the control socket'''
    commandMap = {'request': {'minArguments': 9},
                  'lookup':  {'minArguments': 9},
                  'delete':  {'minArguments': 1}}
    
    def handle(self):
        command = self.request.recv(maxDataSize)
        if command:
            try:
                result = self.execute(command)
                if result is not None:
                    self.request.send(result)
            except CommandError, msg:
                error(msg)
        self.request.close() # this is done by close_request() anyway

    def execute(self, command):
        print "command %s" % command.rstrip()

        ## Split command and check if it is valid
        sessionData = command.split()
        if len(sessionData) < 1:
            raise CommandError, "received empty command '%s'" % command.rstrip()
        cmd = sessionData[0].lower()
        if cmd not in self.commandMap.keys():
            raise CommandError, "received unknown command '%s'" % command.rstrip()

        minArguments = self.commandMap[cmd]['minArguments']
        if len(sessionData) < minArguments + 1:
            suffix = (minArguments!=1 and 's') or ''
            raise CommandError, "the '%s' command needs at least %d argument%s: '%s'" % \
                  (cmd, minArguments, suffix, command.rstrip())

        ## And do the mumbo-jumbo thing
        if cmd == 'request':
            ## Lock before get and release after add to avoid concurency issues
            sessionId = sessionData[1]
            fromDomain = sessionData[4]
            toDomain   = sessionData[6]
            localFrom = (sessionData[5]=='local')
            localTo   = (sessionData[7]=='local')
            domain = (localFrom and fromDomain) or (localTo and toDomain) or None
            proxy = self.proxyFor(domain)
            sessionLock.acquire()
            try:
                try:
                    session = Sessions[sessionId]
                except KeyError:
                    if not proxy:
                        warning("no proxy available for call: %s -> %s" % \
                                (fromDomain, toDomain))
                        return None
                    session = MediaSession(sessionId, proxy)
                    Sessions[sessionId] = session
            finally:
                sessionLock.release()
            return session.forward(command)
        elif cmd == 'lookup':
            sessionId = sessionData[1]
            try:
                session = getSession(sessionId)
            except KeyError:
                warning("trying to lookup session with non-existent id: '%s'" % sessionId)
                return None
            return session.forward(command)
        elif cmd == 'delete':
            sessionId = sessionData[1]
            try:
                session = getSession(sessionId)
            except KeyError:
                return None
            result = session.forward(command)
            session.release()
            return result
        else:
            raise CommandError, "received unknown command: '%s'" % command.rstrip()

    def proxyFor(self, domain):
        '''
        Return the first mediaproxy server that responds from all mediaproxies
        available for the domain. The order in which proxies are checked is based
        on DNS SRV records. RTP traffic will be load balanced between the servers
        at the same priority but with different weights. If some of them are not
        responding their traffic share will be distributed between the other servers
        at the same priority. When none of the servers at a given priority is
        reachable fallback to the next priority where the same rules will apply.
        '''

        def proxyList(domain):
            '''
            Return a list of available mediaproxy servers for the given domain,
            based on DNS SRV records. Traffic will be load balanced between the servers
            at the same priority but with different weights. If some of them are not
            responding their traffic share will be distributed between the other servers
            at the same priority. When none of the servers at a given priority is
            reachable fallback to the next priority where the same rules will apply.
            '''
            mediaproxyrec = "_mediaproxy._tcp.%s" % domain
            answers = DNS.DnsRequest().req(name=mediaproxyrec, qtype='SRV').answers
            if not answers:
                return []
            servers = {}
            for x in answers:
                try:
                    (priority, weight, port, address) = x['data']
                except:
                    continue
                array = servers.setdefault(priority, [])
                array.append((weight, address, port))
            priorities = servers.keys()
            priorities.sort()
            proxyList = []
            for p in priorities:
                all = servers[p] ## all servers at the same priority
                if len(all) == 1:
                    proxyList.append((all[0][1:]))
                else:
                    weights = [x[0] or 1 for x in all]
                    try:
                        totalWeight = sum(weights)
                    except NameError:
                        ## This is for python older than 2.3
                        totalWeight = reduce(lambda x,y: x+y, weights)
                    weights = [int(w*100/totalWeight) for w in weights]
                    array = reduce(lambda x,y: x+y, [weights[i]*[i] for i in
                                                     xrange(len(weights))])
                    while array:
                        index = random.choice(array)
                        proxyList.append(all[index][1:])
                        array = [x for x in array if x!=index]
                    del(array)
            return proxyList

        # start all connections in non-blocking mode the revert to blocking and poll each
        # this way we only wait for 5 seconds no matter how many, instead of n*5 as it is now
        #
        # doesn't work, as we will not be able to use timeouts after reverting to blocking
        # mode. ev start first in blocking mode and the rest in nonblocking mode and check
        # them only if first did timeout.
        if domain is None:
            print "none of the caller or called party is local. will not use mediaproxy."
            return None
        proxies = proxyList(domain)
        ## If there are no mediaproxies defined for a domain use default
        if not proxies:
            # check if to keep this message
            print "domain %s doesn't define any mediaproxy." % domain
            if defaultDomain is None:
                if defaultMediaProxy is not None:
                    # same with this
                    print "will use default mediaproxy for this call."
                return defaultMediaProxy
            proxies = proxyList(defaultDomain)
            if not proxies:
                warning("domain %s, configured as source for default mediaproxies, doesn't define any!" %
                        defaultDomain)
                return None
            domain = defaultDomain
        for proxy in proxies:
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            s.settimeout(5)
            try:
                try:
                    s.connect(proxy)
                except (socket.error, socket.gaierror, socket.herror, socket.timeout):
                    continue
                else:
                    break
            finally:
                s.close()
        else:
            warning("none of the proxies for domain %s have responded" % domain)
            proxy = None
        return proxy


class MediaSession:
    def __init__(self, sessionId, address):
        self._id = sessionId
        self.address = address
        self.startTime = time.time()
        self.expire = self.startTime + sessionTimeout
        if type(address) == type(()):
            self.socktype = socket.AF_INET
            self.address_str = "%s:%s" % address
        else:
            self.socktype = socket.AF_UNIX
            self.address_str = address

    def forward(self, command):
        if self.address is None:
            return None
        s = socket.socket(self.socktype, socket.SOCK_STREAM)
        try: # better check if it is socket.AF_INET. try is bad, because in python2.3 actually works
            # actually it shouldn't matter much. 5 seconds timeout on a unix socket is never met
            s.settimeout(5)
        except AttributeError:
            pass
        result = None
        try:
            try:
                s.connect(self.address)
                s.send(command)
            except (socket.error, socket.gaierror, socket.herror, socket.timeout), why:
                try:               motive = why[1]
                except IndexError: motive = why
                error("couldn't forward command to %s (%s)" % (self.address_str, motive))
                return None
            result = s.recv(maxDataSize)
        finally:
            s.close()
        print "forwarding to mediaproxy on %s: got: '%s'" % (self.address_str, result.rstrip())
        return result

    def release(self):
        sessionLock.acquire()
        try:
            try:
                del(Sessions[self._id])
            except KeyError:
                pass
        finally:
            sessionLock.release()

