#!/usr/bin/env python
# -*- coding: ascii -*-

'''Implementation of a mediaproxy request dispatcher for SER.

 The dispatcher implements an architecture where SER will work with multiple
SERMediaProxy servers, be they local or remote. This can allow both using different
MediaProxy servers for different domains as well as distributing the traffic load on
multiple MediaProxy servers (running on multiple hosts or multiple MediaProxy
processes running on a multi-CPU machine).
 This dispatcher uses dns SRV records to decide where a request should be forwarded.
The dispatcher receives requests from SER's mediaproxy module and decides to which
local or remote MediaProxy to forward. If there are multiple dns SRV records for
a given domain, they will allow both load balancing of RTP traffic as well as
fallbacks in case one is not working.
 After finding the destined MediaProxy server that will handle the request, the
dispatcher will forward the request that came from SER's mediaproxy module to that
server, read back the answer and handle that answer to SER's mediaproxy module.

Copyright 2004 Dan Pascu
'''

## mangle path depending on version (for binary package). also add modules subdir
import sys, os
version = '%d.%d' % sys.version_info[0:2]
sys.path[1:1] = [os.path.normpath((sys.path[0] or '.') + '/' + version)]
sys.path[0:0] = [os.path.normpath((sys.path[0] or '.') + '/modules')]
del os


if __name__ == '__main__':
    import dispatcher
    import daemon
    import sys
    from optparse import OptionParser
    from options import ExtendedOptions

    parser = OptionParser(option_class=ExtendedOptions, version="%prog 1.0")
    parser.add_option("--no-fork", action="store_false", dest="fork", default=1,
                      help="run the process in the foreground")
    parser.add_option("--socket", dest="controlSock", default='/var/run/proxydispatcher.sock',
                      help="control socket (/var/run/proxydispatcher.sock)", 
                      metavar="FILE")
    parser.add_option("--group", dest="group", default='ser',
                      help="allow this group to write to the control socket (ser)", 
                      metavar="GROUP")
    parser.add_option("--proxy", dest="proxy", type="controlport", default='/var/run/mediaproxy.sock',
                      help="default mediaproxy address (/var/run/mediaproxy.sock)", 
                      metavar="Host|File")
    parser.add_option("--pid", dest="pidFile", default='/var/run/proxydispatcher.pid',
                      help="pid file (/var/run/proxydispatcher.pid)", 
                      metavar="FILE")
    parser.add_option("--log", dest="logFile", default=None,
                      help="log to this file too, along with syslog (None)",
                      metavar="FILE")
    (options, args) = parser.parse_args()
    ## Pass the command line options (or the default values) to the rtphandler module
    dispatcher.controlSock = options.controlSock
    dispatcher.controlSockGroup = options.group
    pidFile = options.pidFile
    ## Set default media proxy we will use (or None)
    try:
        proxyname = options.proxy.lower()
    except AttributeError:
        dispatcher.defaultMediaProxy = options.proxy
    else:
        if proxyname == 'none':
            dispatcher.defaultMediaProxy = None
        elif proxyname.startswith('domain://'):
            dispatcher.defaultDomain = options.proxy[9:] or None
        else:
            dispatcher.defaultMediaProxy = options.proxy

    if options.fork:
        try:
            d = daemon.Daemon('proxydispatcher', pidFile, logFile=options.logFile, useSyslog=1)
        except daemon.DaemonError, why:
            print >>sys.stderr, "Mediaproxy dispatcher error: %s. Exiting." % why
            sys.exit(1)

    try:
        ## We can also pass some of the options here
        server = dispatcher.SERCommandHandler(options.controlSock, options.group)

        ## Show some info on startup about the current settings
        print "Listening for commands on local socket `%s'" % options.controlSock
        plural = " is"
        if dispatcher.defaultDomain:
            where = "those defined by domain %s" % dispatcher.defaultDomain
            plural = "s are"
        elif dispatcher.defaultMediaProxy is None:
            where = "disabled.\nDomains without mediaproxy SRV records will not be handled."
        elif type(dispatcher.defaultMediaProxy) is type(()):
            where = "on `%s:%s'" % dispatcher.defaultMediaProxy
        else:
            where = "on `%s'" % dispatcher.defaultMediaProxy
        print "Default mediaproxy server%s %s" % (plural, where)

        dispatcher.run(server)
    except dispatcher.DispatcherError, why:
        print >>sys.stderr, "Mediaproxy dispatcher error: %s." % why
    except KeyboardInterrupt:
        print >>sys.stderr, "Got keyboard interrupt. Exiting."
    except daemon.DaemonQuit, msg:
        print >>sys.stderr, msg

